(* 
   KAIST CS320 2003
   C- language lexer
   Woongsik Choi
   (wschoi@ropas.kaist.ac.kr)
*)   

{
 open Parser
 exception Eof
 exception LexicalError
 fun verbose1 s =  (* (print_string s; print_newline(); s) *) s
 fun verbose2 s =  (* (print_string s; print_newline()) *) ()
 val comment_depth = ref 0
 val keyword_tbl = Hashtbl.create 31
 val _ = List.iter (fn (keyword, tok) => Hashtbl.add keyword_tbl keyword tok)
                   [("true", TRUE),
                    ("false", FALSE),
                    ("not", NOT),
                    ("if", IF),
                    ("then",THEN),
                    ("else",ELSE),
                    ("let", LET),
                    ("in", IN),
                    ("end", END),
	            ("proc", PROC),
                    ("while", WHILE),
                    ("do"   , DO),
                    ("for"  , FOR),
                    ("to"   , TO),
                    ("read" , READ),
                    ("write", WRITE)
                  ] 
} 

val blank = [' ' '\n' '\t']+
    (* KS X 1001 a.k.a. KS C 5601 or euc-kr *)
val hangul = (['\176'-'\200']['\161'-'\254'])+
    (* KS X 1005-1 a.k.a. KS C 5700, unicode, or ISO/IEC 10646-1 *)
val uhangul = (['\172'-'\214']['\000'-'\255']|'\215'['\000'-'\163'])+
val id = (['a'-'z']|hangul|uhangul)
         (hangul|uhangul|['a'-'z' 'A'-'Z' '\'' '0'-'9' '_'])*
val number = ['0'-'9']+

rule start =
 parse blank { start lexbuf }
     | "/*" { comment_depth :=1;
              comment lexbuf;
              start lexbuf }
     | number { NUM (int_of_string (verbose1 (Lexing.lexeme lexbuf))) }
     | id { let val id = verbose1 (Lexing.lexeme lexbuf)
            in Hashtbl.find keyword_tbl id
               handle _ => ID id
            end}
     | "+" {verbose2 "+"; PLUS}
     | "-" {verbose2 "-";MINUS}
     | "*" { verbose2 "*"; STAR}
     | "/" { verbose2 "/"; SLASH}
     | "=" {verbose2 "="; EQUAL}
     | "<" { verbose2 "<"; LB}
     | ">" { verbose2 ">"; RB}
     | "]" { verbose2 "]"; RBLOCK}
     | "[" { verbose2 "["; LBLOCK}
     | ":=" {verbose2 ":="; COLONEQ}
     | ";" { verbose2 ";"; SEMICOLON}
     | "(" { verbose2 "("; LP}
     | ")" { verbose2 ")"; RP}
     | eof { verbose2 "eof"; EOF}
     | _ {raise LexicalError}

and comment = parse
     "/*" {comment_depth := !comment_depth+1; comment lexbuf}
   | "*/" {comment_depth := !comment_depth-1;
           if !comment_depth > 0 then comment lexbuf }
   | eof {raise Eof}
   | _   {comment lexbuf}
