(*
 * SNU 4190.310 Programming Languages (Fall 2008)
 *
 * SM5
 *)

signature SM5 = 
sig
  type cmd = 
         PUSH of obj | POP | STORE | LOAD | JTR of command * command
       | MALLOC | BOX of int | UNBOX of string | BIND of string | UNBIND
       | GET | PUT | CALL | ADD | SUB | MUL | DIV | EQ | LESS | NOT
   and obj = Val of value | Id of string | Fn of string * command
   and value = Z of int | B of bool | L of loc | Unit | R of record
   and record
   and loc
   and command = cmd list

  val empty_command : command

  val print : command -> unit
  val run : command -> unit
end

structure Sm5 : SM5 =
struct
  type cmd = 
         PUSH of obj | POP | STORE | LOAD | JTR of command * command
       | MALLOC | BOX of int | UNBOX of string | BIND of string | UNBIND
       | GET | PUT | CALL | ADD | SUB | MUL | DIV | EQ | LESS | NOT
   and obj = Val of value | Id of string | Fn of string * command
   and svalue = V of value | P of proc | M of map
   and value = Z of int | B of bool | L of loc | Unit | R of record
   and record = map list
   and loc = int * int
   and map = string * svalue
   and proc = string * command * environment
   and stack = svalue list
   and memory = (loc * value) list
   and environment = map list
   and command = cmd list
   and continuation = (command * environment) list
   
  exception RunError of stack * memory * environment * command * continuation
  exception Unbound_id of string
  exception Unbound_loc of int * int
  exception End

  val empty_command = []

  fun (@?) l x = snd (List.find (fn y => x = fst y) l)
  fun fsts l = List.map fst l 
  fun rangecheck l r1 r2 =
  	case l of
	  [] => true
	| h::tl => ((r1 @? h) = (r2 @? h)) andalso (rangecheck tl r1 r2)

  open Format

  fun print_seq f g l =
    case l of
      [] => ()
    | [h] => f h
    | h::t => f h; g h; print_seq f g t
  fun printv v =
    case v of
      Z i => printf "%d" i
    | B b => if b then printf "true" else printf "false"
    | R [] => printf "[]"
    | R (h::t) =>
        let fun pf (x, V(L(l1, l2))) = printf "(%s,<%d,%d>)" x l1 l2
              | pf _ = raise Invalid_argument "non Loc in Record"
        in  printf "["; pf h; List.iter (fn f => printf ", "; pf f) t; 
            printf "]"
        end
    | Unit => printf "()" 
    | L (b,o) => printf "<%d,%d>" b o
  fun printp p = 
    case p of
      Val v => printv v
    | Id x => printf "%s" x
    | Fn(x,c) => printf "@[<1>(%s,@ " x; print c; printf ")@]"
  and printc c = printf "@[";
    (case c of
         PUSH p => printf "push "; printp p
       | POP => printf "pop"
       | STORE => printf "store"
       | LOAD => printf "load"
       | JTR(c1,c2) => 
           printf "@[<5>jtr ("; print c1; printf ",@ ";
           print c2; printf ")@]"
       | MALLOC => printf "malloc"
       | BOX z => printf "box %d" z
       | UNBOX x => printf "unbox %s" x
       | BIND x => printf "bind %s" x
       | UNBIND => printf "unbind"
       | GET => printf "get"
       | PUT => printf "put"
       | CALL => printf "call"
       | ADD => printf "add"
       | SUB => printf "sub"
       | MUL => printf "mul"
       | DIV => printf "div"
       | EQ => printf "eq"
       | LESS => printf "less"
       | NOT => printf "not"); printf "@]"
  and print l =
    printf "@[";
    print_seq printc (fn _ => printf "@ ") l;
    printf "@]";
    print_flush()
    
  val loccount = ref 0
      (* create new loc *)
  fun newl() = loccount++; (!loccount,0)

  fun eval (x as (s,m,e,c,k)) = eval(
     case x of
       (_,_,_,PUSH(Val v)::c,_) => (V v::s, m, e, c, k)
     | (_,_,_,PUSH(Id x)::c, _) => 
         (((e @? x)::s, m, e, c, k) 
          handle Not_found => raise Unbound_id x) 
     | (_,_,_,PUSH(Fn(x,c'))::c,_) => (P(x,c',e)::s, m, e, c, k)
     | (w::s,_,_,POP::c,k) => (s, m, e, c, k)
     | (V(L l)::V v::s,_,_,STORE::c,_) => (s, (l,v)::m, e, c, k)
     | (V(L l)::s,_,_,LOAD::c,_) => 
         ((V(m @? l)::s, m, e, c, k)
           handle Not_found => raise Unbound_loc l)
     | (V(B b)::s,_,_,JTR(c1,c2)::c,_) => 
         (s, m, e, if b then c1@c else c2@c, k)
     | (_,_,_,MALLOC::c,_) => (V(L(newl()))::s, m, e, c, k)
     | (_,_,_,BOX z::c,_) =>
        let fun box b 0 s = V (R b)::s
              | box b z (M m::s) = box (m::b) (z-1) s
              | box _ _ _ = raise RunError (s,m,e,c,k)
        in  (box [] z s,m,e,c,k)
        end
     | (V (R b)::s,_,_,UNBOX x::c,_) =>
        (((b @? x)::s,m,e,c,k)
            handle Not_found => raise Unbound_id x)
     | (w::s,_,_,BIND x::c,_) => (s, m, (x,w)::e, c, k)
     | (_,_,i::e,UNBIND::c,_) => (M i::s, m, e, c, k)
     | (V(L l)::V v::P(x,c',e')::s,_,_,CALL::c,k) =>
         (s, (l,v)::m, (x,V(L l))::e', c', (c,e)::k)
     | (_,_,_,[],(c,e')::k) => (s, m, e', c, k)
     | (_,_,_,GET::c,_) => (V(Z(read_int()))::s, m, e, c, k) 
     | (V(Z z)::s,_,_,PUT::c,_) => 
          print_int z; print_newline(); (s, m, e, c, k)
     | (V(Z z2)::V(Z z1)::s,_,_,ADD::c,_) => (V(Z(z1+z2))::s, m, e, c, k)
     | (V(Z z2)::V(L(l1,z1))::s,_,_,ADD::c,_) => (V(L(l1,z1+z2))::s, m, e, c, k)
     | (V(L(l2,z2))::V(Z z1)::s,_,_,ADD::c,_) => (V(L(l2,z1+z2))::s, m, e, c, k)
     | (V(Z z2)::V(Z z1)::s,_,_,SUB::c,_) => (V(Z(z1-z2))::s, m, e, c, k)
     | (V(Z z2)::V(L(l1,z1))::s,_,_,SUB::c,_) => (V(L(l1,z1-z2))::s, m, e, c, k)
     | (V(L(l2,z2))::V(L(l1,z1))::s,_,_,SUB::c,_) => if l1 = l2 then (V(Z(z1-z2))::s, m, e, c, k)
	 												else raise RunError (s,m,e,c,k)
     | (V(Z z2)::V(Z z1)::s,_,_,MUL::c,_) => (V(Z(z1*z2))::s, m, e, c, k)
     | (V(Z z2)::V(Z z1)::s,_,_,DIV::c,_) => (V(Z(z1/z2))::s, m, e, c, k)
     | (V(Z z2)::V(Z z1)::s,_,_,EQ::c,_) => (V(B(z1=z2))::s, m, e, c, k)
     | (V(B b2)::V(B b1)::s,_,_,EQ::c,_) => (V(B(b1=b2))::s, m, e, c, k)
     | (V(R r2)::V(R r1)::s,_,_,EQ::c,_) =>
       (V(B((fsts r1 = fsts r2) andalso (rangecheck (fsts r1) r1 r2)))::s, m, e, c, k)
    
     | (V Unit::V Unit::s,_,_,EQ::c,_) => (V(B true)::s, m, e, c, k)
     | (V(L(l2,z2))::V(L(l1,z1))::s,_,_,EQ::c,_) => (V(B(l1 = l2 andalso z1 = z2))::s, m, e, c, k)
     | (V _::V _::s,_,_,EQ::c,_) => (V(B false)::s,m,e,c,k) 
     | (V(Z z2)::V(Z z1)::s,_,_,LESS::c,_) => (V(B(z1<z2))::s, m, e, c, k)
     | (V(L(z1,z2))::V(L(l1,l2))::s,_,_,LESS::c,_) => if z1 = l1 then (V(B(l2<z2))::s, m, e, c, k)
	 												else raise RunError (s,m,e,c,k)
     | (V(B b)::s,_,_,NOT::c,_) => (V(B(not b))::s, m, e, c, k)
     | (_,_,_,[],[]) => raise End
     | _ => raise RunError (s,m,e,c,k))

  fun print_error x = printf "SM5 evaluation error: ";
    (case x of
       Unbound_id x => printf "unbound id '%s'.@." x
     | Unbound_loc (l1,l2) => printf "unbound loc (%d,%d).@." l1 l2
     | RunError s => printf "stuck configuration.@."
     | x => raise x);
    print_flush()  

  fun run c = (ignore (eval ([],[],[],c,[]))) 
                handle End => ()
                     | x => print_error x
end
